"""
built-in Operations
$Id: Ops.py,v 1.16 2000/08/29 12:37:57 konrad Exp $
(C) Konrad Anton 2000
"""
import Stack, Stackmachine, math
from Exceptions import *
from types import *
from GMLTypes import *
from PrettyPrint import typename


SM=Stackmachine

# get real or fake scene factory
dummy_raytrace=0

# include println itoa ftoa nudge etc...?
extended_language=1

if dummy_raytrace:
    import DummyFactory
    scfy=DummyFactory.SimpleSceneFactory()
else:
    import RayTrace
    scfy=RayTrace.SimpleSceneFactory()


class Operation:

    def __init__(self, opname, ts, rettype, f, **kwargs):
        """
        define an operation with types ts, rettype rettype
        and lambda-fun f.
        kwargs can contain strict=1 so that the function will not
        be evaluated during optimization.
        """
        self.fun=f
        self.types=ts
        self.nargs=len(ts)
        self.rettype=rettype
        self.opname=opname
        try:     self.strict=kwargs['strict'] # may be eval'd during opt.?
        except:  self.strict=0                    
        
    def assertType(self, propert, theo, argno, opname, coor):
        """
        assert that object theo is of type propert;
        if not, COMPLAIN about type mismatch in Op=opname, Arg=argno,
        line,col=coor
        """
        if type(propert) is ClassType:
            if not isinstance(theo, propert):
                raise GML_TypeError(coor[0],coor[1],
                                    "%s wants %s for arg %d, not %s"\
                                %(opname, typename(propert),
                                  argno, typename(theo)))
        elif type(theo) is not propert:
            raise GML_TypeError(coor[0],coor[1],
                                "%s wants %s for arg %d, not %s (value: %s)"\
                                %(opname, typename(propert),
                                  argno, typename(theo), str(theo)))


    def __call__(self, st, coor):
        "st: machine stack, coor: (line,col)"

        args=st.multipop(self.nargs, coor[0],coor[1])
                                      
        for i in range(self.nargs):
            self.assertType(self.types[i], args[i], i+1, self.opname, coor)

        try:
            result=apply(self.fun, tuple(args))
        except Exception, e:
            raise GML_RuntimeError(coor[0],coor[1],
                                   str(e)+" during call to "+self.opname)
        
        if self.rettype is not NoneType:
            st.push( result )

    def __repr__(self):
        return self.opname



(SCO_NONE, SCO_LIGHT, SCO_OBJECT) = range(3)
#symbolic constants for pseudo-strong typing

class SceneOperation(Operation):
    "an operation that interferes with Cedric's raytracer objects"

    factory = scfy # the scene object factory
        
    def assertType(self, propert, theo, argno, opname, coor):
        """
        assert that object theo is of type propert;
        if not, COMPLAIN about type mismatch in Op=opname, Arg=argno,
        line,col=coor
        """
        fry=self.factory
        
        if propert == SCO_LIGHT:
            if not fry.isLight(theo):
                raise GML_TypeError(coor[0],coor[1],
                                    "Light required for %s arg %d"\
                                    %(opname, argno))

        elif propert == SCO_OBJECT:
            if not fry.isSceneObject(theo):
                raise GML_TypeError(coor[0],coor[1],
                                    "Scene object required for %s arg %d"\
                                    %(opname, argno))

        else: # non-scene-related-type, checked by superclass
            Operation.assertType(self, propert, theo, argno, opname, coor)


def clampfun(x):
    if x <= 0.0:
        return 0.0
    elif x >= 1.0:
        return 1.0
    else: return x

def divifun(x,y):
    "a division that rounds towards zero"
    q,r=divmod(x,y)
    if (q < 0): 
        if r != 0:
            q=q+1 # GML rounds towards zero in negative division, unlike Py
    return q
def modifun(x,y):
    "a modulo suitable for a division that rounds towards zero"
    q,r=divmod(x,y)
    if (q < 0): 
        if r != 0:
            r=r-y # GML rounds towards zero in negative division, unlike Py
    return r

def printfun(s):
    print s,
    
def printlnfun(s):
    print s

def call_later_fun(c,i):
    print "Now, dear closure: what do you associate with 1, 2 and %d?"%i
    print "The closure scratches her head..."
    st=[1,2,i]
    top=c.stack_call(st)
    print "and leaves",st,"on the stack, pointing fiercely at",st[:top]

def debug_render_fun(amb, lights, obj, depth, fov, wid, ht, filen):
    print "Time passes...\n\nMore time passes...\nYou fall asleep."
    print "\nYou dream of ambient light in %s"%str(amb)
    print "and other light coming from:"
    print map(str, lights.items)
    print "\nYou see an object:"
    print str(obj)
    print "\nYou can see up to %d rays at an angle of %f degrees, %d by %d"\
          %(depth,fov,wid,ht)
    print "\nYou wake up and find a new file, %s, on your computer.\n"%filen

def real_render_fun(amb, lights, obj, depth, fov, wid, ht, filen):
    RayTrace.render(amb.asTuple(), lights.asList(), obj, depth, fov,
                    wid, ht, filen)

if dummy_raytrace:
    render_fun=debug_render_fun
else:
    render_fun=real_render_fun

DEGREE_ARC=180.0/math.pi

basic_operations = {
    # BOOLEANS
    # treating them in the parser is easier, but more complex.
    # the optimizer will treat them anyway.
    "true": Operation("true", (), Boolean, lambda : Boolean(1)), 
    "false": Operation("false", (), Boolean, lambda : Boolean(0)),
    
    # NUMBERS
    "addi": Operation("addi",(IntType,IntType),IntType,
                      lambda x,y:x+y),
    "addf": Operation("addf",(FloatType,FloatType),FloatType,
                      lambda x,y:x+y),
    "acos": Operation("acos",(FloatType,), FloatType,
                      lambda x: math.acos(x)*DEGREE_ARC ),
    "asin": Operation("asin",(FloatType,), FloatType,
                      lambda x: math.asin(x)*DEGREE_ARC ),
    "clampf": Operation("clampf",(FloatType,), FloatType,
                        clampfun),
    "cos": Operation("cos",(FloatType,), FloatType,
                     lambda x: math.cos(x/DEGREE_ARC)),
    "divi": Operation("divi",(IntType, IntType), IntType,
                      divifun),
    "divf": Operation("divf",(FloatType, FloatType), FloatType,
                      lambda x,y: x/y),
    "eqi": Operation("eqi", (IntType, IntType), Boolean,
                     lambda x,y: Boolean(x==y)),
    "eqf": Operation("eqf", (FloatType, FloatType), Boolean,
                     lambda x,y: Boolean(x==y)),
    "floor": Operation("floor", (FloatType,), IntType,
                       lambda x: int(math.floor(x))),
    "frac": Operation("frac", (FloatType,), FloatType,
                      lambda x: math.modf(x)[0]),
    "lessi": Operation("lessi", (IntType, IntType), Boolean,
                       lambda x,y: Boolean(x<y)),
    "lessf": Operation("lessf", (FloatType, FloatType), Boolean,
                       lambda x,y: Boolean(x<y)),
    "modi": Operation("modi", (IntType, IntType), IntType,
                      modifun),
    "muli": Operation("muli", (IntType, IntType), IntType,
                      lambda x,y: x*y),
    "mulf": Operation("mulf", (FloatType, FloatType), FloatType,
                      lambda x,y: x*y),
    "negi": Operation("negi", (IntType,), IntType,
                      lambda x: -x),
    "negf": Operation("negf", (FloatType,), FloatType,
                      lambda x: -x),
    "real": Operation("real", (IntType,), FloatType,
                  float),
    "sin": Operation("sin", (FloatType,), FloatType,
                  lambda x: math.sin(x/DEGREE_ARC)),
    "sqrt": Operation("sqrt", (FloatType,), FloatType,
                  math.sqrt),
    "subi": Operation("subi", (IntType, IntType), IntType,
                  lambda x,y:x-y),
    "subf": Operation("subi", (FloatType, FloatType), FloatType,
                  lambda x,y:x-y),

    # POINTS
    "point": Operation("point",(FloatType,FloatType,FloatType), Point,
                       Point),
    "getx": Operation("getx",(Point,),FloatType, lambda p: p.x),
    "gety": Operation("gety",(Point,),FloatType, lambda p: p.y),
    "getz": Operation("getz",(Point,),FloatType, lambda p: p.z),

    # ARRAYS
    "get": Operation("get", (Array,IntType), AnyGMLType,
                     lambda a, i: a.get(i)),
    "length": Operation("length", (Array,), IntType,
                        lambda a: a.length()),

    # OBJECTS
    "cube": SceneOperation("cube", (Closure,), SCO_OBJECT,
                             lambda clo: \
                                 scfy.makeCube(clo)),
    "cylinder": SceneOperation("cylinder", (Closure,), SCO_OBJECT,
                             lambda clo: \
                                 scfy.makeCylinder(clo)),
    "cone": SceneOperation("cone", (Closure,), SCO_OBJECT,
                             lambda clo: \
                                 scfy.makeCone(clo)),
    "sphere": SceneOperation("sphere", (Closure,), SCO_OBJECT,
                             lambda clo: \
                                 scfy.makeSphere(clo)),
    "plane": SceneOperation("plane", (Closure,), SCO_OBJECT,
                             lambda clo: \
                                 scfy.makePlane(clo)),

    # LIGHTS
    "light": SceneOperation("light", (Point, Point), SCO_LIGHT,
                            lambda di,co: \
                                scfy.makeLight(di.asTuple(),co.asTuple())),
    "pointlight": SceneOperation("pointlight", (Point, Point), SCO_LIGHT,
                            lambda di,co: \
                                scfy.makePointLight(di.asTuple(),
                                                    co.asTuple())),
    
    "spotlight": SceneOperation("spotlight",
                                (Point,Point,Point,FloatType,FloatType),
                                SCO_LIGHT,
                                lambda po,di,co,cf,ex:\
                                    scfy.makeSpotLight(po.asTuple(),
                                                       di.asTuple(),
                                                       co.asTuple(),
                                                       cf,ex)),

    # TRANSFORMS
    "translate": SceneOperation("translate",
                                (SCO_OBJECT, FloatType, FloatType, FloatType),
                                SCO_OBJECT,
                                lambda o,x,y,z: \
                                    scfy.applyTranslation(o,x,y,z)),
    "scale": SceneOperation("scale",
                                (SCO_OBJECT, FloatType, FloatType, FloatType),
                                SCO_OBJECT,
                                lambda o,x,y,z: \
                                    scfy.applyScale(o,x,y,z)),
    "uscale": SceneOperation("uscale",
                                (SCO_OBJECT, FloatType),
                                SCO_OBJECT,
                                lambda o,s: \
                                    scfy.applyUScale(o,s)),
    "rotatex": SceneOperation("rotatex",
                                (SCO_OBJECT, FloatType),
                                SCO_OBJECT,
                                lambda o,phi: \
                                    scfy.applyRotateX(o,phi)),

    "rotatey": SceneOperation("rotatey",
                                (SCO_OBJECT, FloatType),
                                SCO_OBJECT,
                                lambda o,phi: \
                                    scfy.applyRotateY(o,phi)),

    "rotatez": SceneOperation("rotatez",
                                (SCO_OBJECT, FloatType),
                                SCO_OBJECT,
                                lambda o,phi: \
                                    scfy.applyRotateZ(o,phi)),

    # CSG
    "union": SceneOperation("union",
                            (SCO_OBJECT, SCO_OBJECT),
                            SCO_OBJECT,
                            lambda a,b: scfy.makeUnion(a,b)),
    "intersect": SceneOperation("intersect",
                            (SCO_OBJECT, SCO_OBJECT),
                            SCO_OBJECT,
                            lambda a,b: scfy.makeIntersect(a,b)),
    "difference": SceneOperation("difference",
                            (SCO_OBJECT, SCO_OBJECT),
                            SCO_OBJECT,
                            lambda a,b: scfy.makeDifference(a,b)),

    # RENDER
    "render": SceneOperation("render",
                             (Point, Array, SCO_OBJECT, IntType,
                              FloatType, IntType, IntType, StringType),
                             NoneType,
                             render_fun,
                             strict=1),

    #DEBUGGING optional
}

ext_operations={}

exts = {
        "itoa": Operation("itoa", (IntType,), StringType,
                          str),
        "ftoa": Operation("ftoa", (FloatType,), StringType,
                          str),
        "ptoa": Operation("ptoa", (Point,), StringType,
                          str),
        "btoa": Operation("btoa", (Boolean,), StringType,
                          str),
        "print": Operation("print", (StringType,), NoneType,
                           printfun, strict=1),
        "println": Operation("println", (StringType,), NoneType,
                           printlnfun, strict=1),
        "call_later": Operation("call_later", (Closure, IntType),
                                         NoneType,
                                         call_later_fun, strict=1),
        "nudge": SceneOperation("nudge", (SCO_OBJECT, StringType), NoneType,
                                lambda sco, s: sco.nudge(s), strict=1),
        "turnon": SceneOperation("turnon", (SCO_LIGHT,), NoneType,
                                 lambda scl: scl.nudge(), strict=1)
    }

ext_operations=basic_operations.copy()
for n,o in exts.items():
    ext_operations[n]=o

def setExtendedLanguage( what ):
    global operations
    extended_language=what
    if what:
        operations=ext_operations
    else:
        operations=basic_operations

#match global setting
setExtendedLanguage(extended_language)
