"""
A stackmachine bytecode optimizer
$Id: Optimizer.py,v 1.5 2000/08/29 12:37:58 konrad Exp $
(C) 2000 Konrad Anton
"""

from Stackmachine import *
import PrettyPrint
from types import *

class Optimizer(Stackmachine):

    def __init__(self, stack=None, symbols=None, **kwargs):
        Stackmachine.__init__(self,stack,symbols)
        self.constitems=0
        if kwargs.has_key("verbose"):
            self.verbose=kwargs["verbose"]
        else:
            self.verbose=0

    # states of accessing a symbol
    (SA_IGNORE,   # not at all accessed
     SA_BINDC,    # bound to a known constant
     SA_BINDV,    # bound to something unknown 
     SA_PUSHV,    # pushing its outside-defined value
    )=range(4)
    accessStateNames=["SA_IGNORE","SA_BINDC","SA_BINDV","SA_PUSHV"]
    def newAccessState(self, asyms, sym, change):
        """
        does transitions between access states
        param: oldstate
        returns: newstate
        """
        try: old = asyms[sym] 
        except: old=self.SA_IGNORE #undefined symbols are undefined
        
        if old==self.SA_IGNORE:
            return change
        elif old==self.SA_BINDC:
            if change==self.SA_BINDV: return change
            else: return old
        else: return old

    def genPushcInstrs(self, coor):
        """
        used when the stack top becomes uncertain. The known-until-then
        stack top elements have to be written out in PUSHC instructions;
        they haven't been written out that way because they might have
        been optimized away.

        coor is the editor coordinate this should be attributed to

        returns a list of pushc instructions and resets the constitems
        counter
        """
        ilist = []

        for stit in self.stack.items[self.stack.top-self.constitems\
                                     :self.stack.top]:
            ilist.append((coor, PUSHC, stit))

        self.constitems=0

        return ilist
    
    def run(self, ilist, asyms=None):
        """
        optimizes a sequence of instructions.
        Params:
         ilist: the instrs
         asyms: a hash symname->Access State constant, won't be modified
        returning a (new ilist, changed asyms) tuple
        """
        V=self.verbose
        st=self.stack
        syms=self.symbols
        oinsts=[]
        if asyms is None: asyms={}
        else:             asyms=asyms.copy()

        # reset SA_BINDV marks on import
        # or we'll barf on { x /y 10 /c { y c addi } }, because
        # y will be set to SA_BINDV in the outer closure, but
        # is nicked from the inner closure.
        # c will be set to SA_BINDC, which is perfect since it is
        # guaranteed to contain 10.
        for sym in asyms.keys(): 
            if asyms[sym]==self.SA_BINDV:
                if V: print "un-nicking",sym
                asyms[sym]=self.SA_IGNORE
                
        
        self.constitems=0 # how many known constant items are on stack top

        for instr in ilist:
            coor, opcode, arg = instr
            line, col=coor

##            print "\nReading:",PrettyPrint.printSMInstr(instr)
##            print "Optimizer stack: ", st.items[:st.top]
##            print "so far",len(oinsts),"insts generated"

            #this could be done more reusably with a central
            # bytecode dispatcher, but that's slow.
            if opcode == NOP: 
                pass
            
            elif opcode==PUSHC:
                st.push(arg)
                self.constitems = self.constitems + 1

            elif opcode==PUSHV:
                # push nothing here
                if asyms[arg] == self.SA_BINDC : # known constant value
                    if V: print "PUSHV",arg,"SA_BINDC"
                    st.push(syms[arg])
                    self.constitems = self.constitems + 1
                else: # no known value
                    if V: print "PUSHV",arg,"SA_PUSHV"
                    asyms[arg]=self.newAccessState(asyms, arg, self.SA_PUSHV)
                    oinsts.extend(self.genPushcInstrs(coor))
                    oinsts.append(instr)
                            
            elif opcode==BIND:
                if self.constitems >= 1:
                    syms[arg]=st.pop(line,col)
                    self.constitems = self.constitems-1
                    asyms[arg] = self.newAccessState(asyms, arg, self.SA_BINDC)
                else: # nothing known on the stack
                    asyms[arg] = self.newAccessState(asyms, arg, self.SA_BINDV)
                    oinsts.append(instr)
                
            elif opcode==OPERATOR:
                if self.constitems >= arg.nargs and not arg.strict:
                    if self.verbose:
                        if V: print "Shortcutting: ", arg.opname,\
                              "-- stack top surely",\
                              st.items[st.top-self.constitems:st.top]
                    ret=arg(st, coor) # shortcut evaluation
                    self.constitems = self.constitems - arg.nargs
                    if arg.rettype is not NoneType:
                        # The result is on the stack and known
                        self.constitems = self.constitems + 1
                        
                else: # the operator touches at least one unknown item
                    oinsts.extend(self.genPushcInstrs(coor))
                    oinsts.append(instr)
                
            elif opcode==APPLY:
                # we do not optimize that at all -- yet.
                oinsts.extend(self.genPushcInstrs(coor))
                oinsts.append(instr)
                    
                    
            elif opcode==CLOSURE:
                if V: print "Trying to optimize closure in", coor
                opti=Optimizer(None, syms.copy(), verbose=self.verbose)
                
                clo_constant = 1 # assume it's a constant closure, i.e.
                                 # it does not depend on outer scope
                try:
                    new_clo_ilist, clo_asyms = opti.run(arg, asyms)
                    if V: print "closure in",coor,"returned asyms..."
                    for symname in clo_asyms.keys():
                        if V: print "ASYM ", symname,\
                           self.accessStateNames[clo_asyms[symname]]
                        if clo_asyms[symname] == self.SA_PUSHV:
                            # closure nicked a name from outer scope
                            asyms[symname] = self.SA_PUSHV
                            clo_constant = 0
                
                except:
                    if V: print "closure",coor,"would cause error.",\
                                "Not optimizing."
                    clo_constant = 0
                    new_clo_ilist=arg
                            
                if clo_constant: # a constant closure, PUSHC it
                    new_closure=Closure(syms,new_clo_ilist)
                    if V: print "creating instant closure",\
                                str(new_closure),"in",coor
                    st.push( new_closure )
                    self.constitems = self.constitems+1
                                    
                else: # a closure that has to be evaluated afresh each time
                    if V: print "creating slow closure in",coor
                    oinsts.extend(self.genPushcInstrs(coor))
                    oinsts.append( (coor, CLOSURE, new_clo_ilist) )



            elif opcode==ARRAY: 
                opti=Optimizer(None, syms.copy(), verbose=self.verbose)
                new_arr_ilist, arr_asyms = opti.run(arg, asyms)
                
                arr_constant = 1  # true iff arr does not depend on names
                                  # from its outer scope
                                  
                for symname in arr_asyms.keys():
                        if arr_asyms[symname] == self.SA_PUSHV:
                            # array nicked a name from outer scope
                            asyms[symname] = self.SA_PUSHV
                            arr_constant = 0 
                            
                if arr_constant: # a constant array, really evaluate it
                    if V: print "Freezing array in line %d col %d"%(coor)
                    new_array=Array(syms,new_arr_ilist)
                    st.push( new_array )
                    self.constitems = self.constitems+1
                                    
                else: # an array that has to be evaluated afresh each time
                    oinsts.extend(self.genPushcInstrs(coor))
                    oinsts.append( (coor, ARRAY, new_arr_ilist) )
                
            elif opcode==ITE:
                # no optimization so far
                oinsts.extend(self.genPushcInstrs(coor))
                oinsts.append(instr)
                
                        
            else:
                print "Opcode %s not supported"%instrname[opcode]

        #finally flush all those constants we earned lately...
        if len(ilist) > 0:
            oinsts.extend(self.genPushcInstrs(coor))
        
        return (oinsts, asyms)
