"""
A simple GML parser without external lex/yacc-like tools
$Id: Parser.py,v 1.9 2000/08/29 12:37:59 konrad Exp $
(C) 2000 Konrad Anton
"""

import re,string
import Stackmachine
import Ops
from Exceptions import *
import PrettyPrint

SM=Stackmachine

(NOTHING, SYMBOL, BINDER, OPENBRACE, CLOSEBRACE,
OPENBRACKET, CLOSEBRACKET, INTLIT, FLOATLIT, STRINGLIT,
EOF) = range(11)

class EOFException(Exception):
    pass


class GML_Lexer:
    """
    Delivers tokens from a readline()-able GML source
    """
    toktypes=[
        (NOTHING, re.compile(r'\s+')),
        (NOTHING, re.compile(r'%.*')),
        (SYMBOL, re.compile(r'[a-zA-Z][a-zA-Z0-9_-]*')),
        (BINDER, re.compile(r'/[a-zA-Z][a-zA-Z0-9_-]*')),
        (OPENBRACE, re.compile(r'{')),
        (CLOSEBRACE, re.compile(r'}')),
        (OPENBRACKET, re.compile(r'\[')),
        (CLOSEBRACKET, re.compile(r'\]')),
        (FLOATLIT, re.compile(r'-?\d+(\.\d+([eE]-?\d+)?|([eE]-?\d+))')),
        (INTLIT, re.compile(r'-?\d+(?![\.eE])')),
        (STRINGLIT, re.compile(r'"[^"]*"'))
        ]
    
    def __init__(self, source):
        self.source=source
        self.lineno=0 # the current line number, 0-based
        self.curline="" # The contents of the line we're working on
        self.colno=0 # index of the next-to-be-matched character

    def fetchNextLine(self):
        self.lineno = self.lineno+1
        aline=self.source.readline()
        if aline:
            if aline[-1]=='\n':
                self.curline=aline[:-1] #strip \n
            else:
                self.curline=aline
            self.colno=0
            return 1
        else: #EOF
            raise EOFException()
        
    def getNextToken(self):
        """
        get token in form of (tokid, content, line, col)
        (line, col) in editor coordinates
        returns (EOF, None, line, 0) on EOF
        """
        rettokenid=NOTHING
        try:
            
            while rettokenid == NOTHING:
                while len(self.curline)-self.colno <= 0:
                    self.fetchNextLine()
                foundp=None
                for tid, regexp in self.toktypes:
                    #print "MATCHING %d %s"%(tid,self.curline[self.colno:])
                    m=regexp.match(self.curline, self.colno)
                    if m:
                        #print "BINGO"
                        foundp=1
                        colno_found=self.colno
                        self.colno=m.end()
                        if tid != NOTHING:
                            return (tid, m.group(0),
                                    self.lineno, colno_found+1)
                if not foundp:
                    raise GML_SyntaxError(self.lineno+1, self.colno+1,
                                          "Unknown Token")
                        
        except EOFException:
            return (EOF, None, self.lineno, 0)


def test_lexer(filename):
    fo=open(filename, "rt")
    le=GML_Lexer(fo)
    while 1:
        (ty,cont,li,col) = le.getNextToken()
        print ty, cont, li, col
        if ty == EOF: break
    fo.close()



class GML_Parser:
    """
    takes a lexer's token stream and gives a stack machine program
    """

    def __init__(self, lexer):
        self.lexer=lexer

    def getProgram(self, closing=EOF, outerscope_syms={}):
        """
          produces a stack machine program from the rest of the tokens
          closing can contain a token type that will finish this parsing
          process and return the program generated so far.
          outerscope_syms contains true for every user-defined symbol 
            (not to be modified)
        """  
        prog=[]
        syms=outerscope_syms.copy() # contains true for user-defined symbols

        while 1:
            ttype, content, line, column = self.lexer.getNextToken()
            lc=(line, column)
            if ttype==closing:
                break
            
            elif ttype==SYMBOL:
                if syms.has_key(content): # user-def symbol
                    prog.append( (lc, SM.PUSHV, content) )
                elif Ops.operations.has_key(content): # intrinsic
                    #todo: static type checking
                    prog.append( (lc, SM.OPERATOR, Ops.operations[content]) )
                elif content=="apply":
                    prog.append( (lc, SM.APPLY, None) )
                elif content=="if":
                    prog.append( (lc, SM.ITE, None) )
                #elif content=="true":
                #    prog.append( (lc, SM.PUSHC, Ops.Boolean(1)))
                #elif content=="false":
                #    prog.append( (lc, SM.PUSHC, Ops.Boolean(0)))
                else:
                    raise GML_NameError(line,column,"%s not defined"%content)

            elif ttype==INTLIT:
                try:
                    prog.append( (lc, SM.PUSHC, string.atoi(content)))
                except ValueError, e:
                    raise GML_SyntaxError(line,column,"in Int literal: "\
                                          +str(e))

            elif ttype==FLOATLIT:
                try:
                    prog.append( (lc,
                                  SM.PUSHC, string.atof(content)))
                except ValueError, e:
                    raise GML_SyntaxError(line,column,"in Float literal: "\
                                          +str(e))

            elif ttype==STRINGLIT:
                prog.append( (lc, SM.PUSHC, content[1:-1]))
                
            elif ttype==BINDER:
                sym=content[1:]
                if Ops.operations.has_key(sym):
                    raise GML_NameError(line,column,"%s is reserved"%sym)
                syms[sym]=1
                prog.append( (lc,SM.BIND, sym) ) # todo: prevent /true
                
            elif ttype==OPENBRACE:
                prog.append( (lc,SM.CLOSURE,
                              self.getProgram(CLOSEBRACE, syms)))
                
            elif ttype==OPENBRACKET:
                prog.append( (lc,SM.ARRAY,
                              self.getProgram(CLOSEBRACKET, syms)))

            elif ttype ==CLOSEBRACE:
                raise GML_SyntaxError(line,column,"unexpected '}' ")
            elif ttype ==CLOSEBRACKET:
                raise GML_SyntaxError(line,column,"unexpected ']' ")
            elif ttype ==EOF:
                raise GML_SyntaxError(line,column,"unexpected EOF ")
        return prog
    
