"""
A stackmachine.
$Id: Stackmachine.py,v 1.9 2000/08/29 12:38:00 konrad Exp $
(C) 2000 Konrad Anton

It contains a stack of values and a symbol table.

This stack machine operates on list of commands of the shape:
(SRCLOC, INSTR, ARG)

where INSTR is a numeric constant and ARG is INSTR-dependent data.
SRCLOC is of the form (line, col).

INSTR:       ARG
  PUSHC      value                     push a constant
  PUSHV      name                      push a variable
  BIND       name                      bind the stack top to a variable
  APPLY      None                      applies the stack top
  OPERATOR   function                  the the built-in operator
  ARRAY      ilist                     makes a fresh array 
  CLOSURE    ilist                     makes a closure
  ITE        None                      does if-then-else (->task)

value  a corresponding value as a python int or python string or 
       a python class. GML Typing can be derived from Python typing.
name   is a variable name to look up in the symbol table
function is a python callable that takes a stack object and a (line,col)
        tuple
"""

import Ops, Stack
from PrettyPrint import *
from GMLTypes import *
from Exceptions import *


#instructions
(NOP, PUSHC, PUSHV, BIND, APPLY,
 OPERATOR, ARRAY, CLOSURE, ITE) = range(9)
    
instrname=["NOP","PUSHC","PUSHV","BIND","APPLY","OPERATOR","ARRAY",
           "CLOSURE","ITE"]


class Stackmachine:
    """A stackmachine bytecode interpreter"""
    
    verbose=0 #whether all stack machines should be made verbose

    def __init__(self, stack=None, symbols=None, **kwargs):
        "takes a stack and a symbol table, both of which may be modified"
        self.setStackAndSymbols(stack or Stack.Stack(), symbols or {})
        try:
            Stackmachine.verbose=kwargs['verbose']
        except:
            pass

    def setStackAndSymbols(self, stack, symbols):
        self.stack=stack
        self.symbols=symbols

    def run(self, ilist):
        st=self.stack
        syms=self.symbols

        if self.verbose: pp=InstrListPrinter(ilist)
        
        for coor, opcode, arg in ilist:
            line, col=coor

            if self.verbose:
                print "exec: ",
                pp.printInstr( (coor,opcode,arg) )

            if opcode == NOP:
                pass
            
            elif opcode==PUSHC:
                st.push(arg)
                
            elif opcode==PUSHV:
                st.push(syms[arg]) # name checking here is up to the parser
                
            elif opcode==BIND:
                x=st.pop(line,col)
                syms[arg]=x
                
            elif opcode==OPERATOR:
                apply( arg, (st, coor))
                
            elif opcode==APPLY:
                f=st.pop(line,col)
                if not isinstance(f, Closure):
                    raise GML_TypeError(line,col,
                                        "apply of a %s"%typename(f))
                else:
                    f.run(st)
                    
                    
            elif opcode==CLOSURE:
                st.push(Closure(syms, arg))

            elif opcode==ARRAY:
                st.push(Array(syms, arg))
                
            elif opcode==ITE:
                (cond, thenform, elseform) = st.multipop(3, line, col)
                if not isinstance(cond, Boolean):
                    raise GML_TypeError(line,col,
                                        "if-condition must be Boolean")
                else:
                    if not isinstance(thenform, Closure) \
                       or not isinstance(elseform, Closure):
                        raise GML_TypeError(line,col,"if wants two closures")
                    else:
                        if cond.value:
                            thenform.run(st)
                        else:
                            elseform.run(st)
                        
            else:
                print "Opcode %s not supported"%instrname[opcode]
                
    
